"""
Test running module in CLI.
"""

from unittest import TestCase, main
from unittest.mock import patch, call, MagicMock

from bear.common import Hasher
from bear.__main__ import run, set_log_levels


class MainCase(TestCase):
    """
    Testing CLI parameters and __main__.py.
    """

    def test_global_log_level(self):
        """
        Test setting log level for all Bear loggers.
        """
        mock_logger = MagicMock()
        patch_get = patch(
            'bear.__main__.logging.getLogger',
            return_value=mock_logger
        )
        with patch_get as get:
            set_log_levels(9000)
            self.assertEqual(get.mock_calls, [
                call('bear.hashing'), call('bear'),
                call('bear.output'), call('bear.__main__')
            ])
            self.assertEqual(mock_logger.mock_calls, [call.setLevel(9000)] * 4)

    @staticmethod
    def test_hash_file():
        """
        Test calling hashing function for a file.
        """
        target = 'testpath'
        hash_file = patch('bear.__main__.hash_file')
        sysargv = patch('sys.argv', [__name__, '--files', target])
        with sysargv, hash_file as mocked:
            run()
            mocked.assert_called_once_with(path=target, hasher=Hasher.MD5)

    @staticmethod
    def test_find_files():
        """
        Test calling function for collecting all files in a folder.
        """
        target = 'testpath'
        find_file = patch('bear.__main__.find_files')
        sysargv = patch('sys.argv', [__name__, '--traverse', target])
        with sysargv, find_file as mocked:
            run()
            mocked.assert_called_once_with(target)

    def test_hash_folder(self):
        """
        Test collecting all files in a folder and calling hashing
        function on them.
        """
        target = 'testpath'
        sysv = patch('sys.argv', [__name__, '--hash', target])
        hfile = patch('bear.__main__.hash_files')
        ffile_ret = [None]
        ffile = patch('bear.__main__.find_files', return_value=ffile_ret)
        fhash = patch('bear.__main__.filter_files')

        with sysv, hfile as m_hash, ffile as m_find, fhash as m_filter:
            run()
            m_find.assert_called_once_with(target)
            m_hash.assert_called_once_with(
                files=m_find.return_value, hasher=Hasher.MD5
            )
            self.assertEqual(m_find.return_value, ffile_ret)
            m_filter.assert_called_once_with(m_hash.return_value)


if __name__ == '__main__':
    main()



"""
Test file and folder manipulation.
"""

from unittest import TestCase, main
from unittest.mock import patch, MagicMock, call
from os.path import join, basename

from ensure import ensure_annotations

from bear.hashing import hash_files
from bear.common import ignore_append, Hasher
from bear.output import (
    find_files, filter_files, find_duplicates, output_duplicates
)


class HashCase(TestCase):
    """
    Test file and folder manipulation functions.
    """

    @staticmethod
    def test_find_files_nonexisting():
        """
        Test calling bear.find_files on non-existing folder.
        """
        with patch('bear.output.LOG.critical') as critical:
            find_files('_' * 30)
            critical.assert_called_once()

    def test_find_files(self):
        """
        Test hashing a string.
        """
        mocked_walk = [(
            'bear', [
                'emptyfolder', 'tests', '__pycache__'
            ], [
                '__main__.py', '__init__.py'
            ]
        ), (
            'bear/emptyfolder', [], []
        ), (
            'bear/tests', ['__pycache__'], [
                'test_files.py', 'test_main.py',
                '__init__.py', 'test_hash.py'
            ]
        ), (
            'bear/tests/__pycache__', [], [
                'test_files.cpython-37.pyc', 'test_hash.cpython-37.pyc',
                '__init__.cpython-37.pyc', 'test_main.cpython-37.pyc'
            ]
        ), (
            'bear/__pycache__', [], [
                '__init__.cpython-37.pyc', '__main__.cpython-37.pyc'
            ]
        )]

        expected = [
            join(name, fname)
            for name, folder, files in mocked_walk
            for fname in files
        ]

        patch_walk = patch('bear.output.walk', return_value=mocked_walk)
        with patch_walk, patch('bear.output.exists', return_value=True):
            self.assertEqual(find_files('_' * 30), expected)

    def test_filter_files(self):
        """
        Test removing duplicates from dictionary of hashes + files.
        """
        hashes = {
            'empty': [], 'original': ['file'],
            'orig+dup': ['file', 'duplicate'],
            'orig+2xdup': ['file', 'duplicate2']
        }
        self.assertEqual(
            filter_files(hashes), dict(
                item for item in hashes.items()
                if item[0] not in ('empty', 'original')
            )
        )

    @staticmethod
    def test_ignore_append():
        """
        Test writing ignored path to a file.
        """
        pid = 12345
        patch_pid = patch('bear.common.getpid', return_value=pid)

        file_obj = MagicMock()
        mocked_open_inst = MagicMock(
            __enter__=MagicMock(return_value=file_obj)
        )
        patch_open = patch(
            'builtins.open', return_value=mocked_open_inst
        )

        with patch_pid, patch_open as mocked_open:
            ignore_append('lalala')

            mocked_open.assert_called_once_with(f'{pid}_ignored.txt', 'a')

            file_obj.write.assert_has_calls([call('lalala'), call('\n')])

    def test_hash_files_memoryerror(self):
        """
        Test hashing a list of files returning MemoryError on file.read().
        """
        def raise_memory_error(*_):
            raise MemoryError()

        patch_log = patch('bear.hashing.LOG.critical')
        patch_hash = patch('builtins.open', raise_memory_error)
        patch_ignore = patch('bear.hashing.ignore_append')
        with patch_hash, patch_log as critical, patch_ignore as ignore:
            inp = ['Nonefile', 'someNone']
            hash_files(inp, Hasher.MD5)
            self.assertEqual(ignore.mock_calls, [call(inp[0]), call(inp[1])])
            self.assertEqual(len(critical.mock_calls), len(inp))

    def test_hash_files_permissionerror(self):
        """
        Test hashing a list of files returning PermissionError on file.read().
        """
        def raise_permission_error(*_):
            raise PermissionError()

        patch_log = patch('bear.hashing.LOG.critical')
        patch_hash = patch('builtins.open', raise_permission_error)
        patch_ignore = patch('bear.hashing.ignore_append')
        with patch_hash, patch_log as critical, patch_ignore as ignore:
            inp = ['lalala', 'somefile']
            hash_files(inp, Hasher.MD5)
            self.assertEqual(ignore.mock_calls, [call(inp[0]), call(inp[1])])
            self.assertEqual(len(critical.mock_calls), len(inp))

    def test_hash_files_filenotfound(self):
        """
        Test hashing a list of non-existing files.
        """
        patch_log = patch('bear.hashing.LOG.warning')
        patch_ignore = patch('bear.hashing.ignore_append')
        with patch_log as warning, patch_ignore as ignore:
            inp = ['does_not_exist_123', 'does_not_exist_456']
            hash_files(inp, Hasher.MD5)
            self.assertEqual(ignore.mock_calls, [call(inp[0]), call(inp[1])])
            self.assertEqual(len(warning.mock_calls), len(inp))

    def test_hash_files_oserror(self):
        """
        Test hashing a list of files returning OSError on file.read().
        """
        def raise_os_error(*_):
            raise OSError()

        patch_log = patch('bear.hashing.LOG.critical')
        patch_hash = patch('builtins.open', raise_os_error)
        patch_ignore = patch('bear.hashing.ignore_append')
        with patch_hash, patch_log as critical, patch_ignore as ignore:
            inp = ['lalala', 'somefile']
            hash_files(inp, Hasher.MD5)
            self.assertEqual(ignore.mock_calls, [call(inp[0]), call(inp[1])])
            self.assertEqual(len(critical.mock_calls), len(inp))

    def test_hash_files(self):
        """
        Test hashing a list of files returning hashes + paths.
        """

        file_hash = {
            'first': '123', 'second': '456',
            'original': '789', 'duplicate': '789'
        }

        # pylint: disable=unused-argument
        # is actually used for kwargs comparison which is more important
        @ensure_annotations
        def _side_effect(path: str, hasher: Hasher):
            return file_hash[path]

        with patch('bear.hashing.hash_file', side_effect=_side_effect):
            inp = ['first', 'second', 'original', 'duplicate']
            out = hash_files(files=inp, hasher=Hasher.MD5)

        expected = {}
        for key, val in file_hash.items():
            if val not in expected:
                expected[val] = [key]
            else:
                expected[val].extend([key])
        self.assertEqual(out, expected)

    def test_find_duplicates_jobs(self):
        """
        Test finding duplicates using correct job count from parameter.
        """
        max_cpu = 666
        patch_cpu = patch('bear.output.cpu_count', return_value=max_cpu)
        patch_pool = patch('bear.output.Pool')

        # because partial() != partial() in mock calls!
        # partially fun, partially headache -_-'
        fun_part = patch('bear.output.partial')

        # pylint: disable=confusing-with-statement
        with patch_cpu, patch_pool as pool, fun_part as partial_fun:
            self.assertEqual(find_duplicates(
                folders=[], processes=1, hasher=Hasher.MD5
            ), {})
            self.assertEqual(find_duplicates(
                folders=[], processes=0, hasher=Hasher.MD5
            ), {})
            self.assertEqual(find_duplicates(
                folders=[], processes=4, hasher=Hasher.MD5
            ), {})

            self.assertEqual([
                # remove __iter__() calls because those return a tuple
                # iterator instead of call().__enter__().map().__iter__()
                item for item in pool.mock_calls if '__iter__' not in str(item)
            ], [
                call(processes=1),
                call().__enter__(),
                call().__enter__().map(partial_fun(
                    hash_files, hasher=Hasher.MD5
                ), []),
                call().__exit__(None, None, None),

                call(processes=666),
                call().__enter__(),
                call().__enter__().map(partial_fun(
                    hash_files, hasher=Hasher.MD5
                ), []),
                call().__exit__(None, None, None),

                call(processes=4),
                call().__enter__(),
                call().__enter__().map(partial_fun(
                    hash_files, hasher=Hasher.MD5
                ), []),
                call().__exit__(None, None, None),
            ])

    def test_find_duplicates_chunks(self):
        """
        Test chunking list of files for multiple processes.
        """
        patch_pool = patch('bear.output.Pool')

        files = [str(num) for num in range(15)]

        def side_effect(folder):
            return {
                'a': files[:5],
                'b': files[5:10],
                'c': files[10:]
            }[basename(folder)]

        patch_find_files = patch(
            'bear.output.find_files', side_effect=side_effect
        )

        # because partial() != partial() in mock calls!
        # partially fun, partially headache -_-'
        fun_part = patch('bear.output.partial')

        # pylint: disable=confusing-with-statement
        with patch_pool as pool, patch_find_files, fun_part as partial_fun:
            self.assertEqual(find_duplicates(
                folders=['a', 'b', 'c'], processes=1, hasher=Hasher.MD5
            ), {})
            self.assertEqual(find_duplicates(
                folders=['a', 'b', 'c'], processes=3, hasher=Hasher.MD5
            ), {})
            self.assertEqual(find_duplicates(
                folders=['a', 'b', 'c'], processes=4, hasher=Hasher.MD5
            ), {})

            self.assertEqual([
                # remove __iter__() calls because those return a tuple
                # iterator instead of call().__enter__().map().__iter__()
                item for item in pool.mock_calls if '__iter__' not in str(item)
            ], [
                call(processes=1),
                call().__enter__(),
                # all files to single process
                call().__enter__().map(partial_fun(
                    hash_files, hasher=Hasher.MD5
                ), [files]),
                call().__exit__(None, None, None),

                call(processes=3),
                call().__enter__(),
                # files chunked len / processors -> 15 // 3 chunks
                call().__enter__().map(partial_fun(
                    hash_files, hasher=Hasher.MD5
                ), [
                    files[0:5], files[5:10], files[10:15]
                ]),
                call().__exit__(None, None, None),

                call(processes=4),
                call().__enter__(),
                # files chunked len / processors -> 15 // 4 chunks
                call().__enter__().map(partial_fun(
                    hash_files, hasher=Hasher.MD5
                ), [
                    files[0:3], files[3:6], files[6:9], files[9:12], files[12:]
                ]),
                call().__exit__(None, None, None),
            ])

    def test_find_duplicates_join(self):
        """
        Test joining duplicates from multiple jobs.
        """
        mock_pool = MagicMock(**{
            '__enter__.return_value.map.return_value': [
                {'123': ['original'], '456': ['ori', 'dupli']},
                {'123': ['duplicate']},
                {'789': ['original'], '012': ['orig', 'dup']}
            ]
        })
        patch_pool = patch('bear.output.Pool', return_value=mock_pool)
        with patch_pool:
            # 789 is single-file original, ignored in filter_files()
            self.assertEqual(find_duplicates(
                folders=[], processes=2, hasher=Hasher.MD5
            ), {
                '012': ['orig', 'dup'],
                '123': ['original', 'duplicate'],
                '456': ['ori', 'dupli']
            })

    @staticmethod
    def test_find_duplicates_output():
        """
        Test outputing duplicates to file.
        """
        hashes = {
            '012': ['orig', 'dup'],
            b'123': ['original', 'duplicate'],
            '456': ['ori', 'dupli']
        }

        file_obj = MagicMock()
        mocked_open_inst = MagicMock(
            __enter__=MagicMock(return_value=file_obj)
        )
        patch_open = patch(
            'builtins.open', return_value=mocked_open_inst
        )
        now = '12345'
        patch_now = patch(
            'bear.output.datetime', now=MagicMock(return_value=now)
        )

        def assert_write_hashes(mock_obj):
            calls = []
            for key, value in hashes.items():
                calls += [
                    call(
                        key.encode('utf-8')
                        if not isinstance(key, bytes) else key
                    ), call(b'\n')
                ]
                calls += [
                    call(b'\t' + val.encode('utf-8') + b'\n')
                    for val in value
                ]
                calls += [call(b'\n\n')]
            mock_obj.assert_has_calls(calls)

        with patch_now, patch_open as mocked_open:
            output_duplicates(hashes)
            mocked_open.assert_called_once_with(f'{now}_duplicates.txt', 'wb')
            assert_write_hashes(file_obj.write)

        with patch_now, patch_open as mocked_open:
            out = 'custom.txt'
            output_duplicates(hashes, out=out)
            mocked_open.assert_called_once_with(out, 'wb')
            assert_write_hashes(file_obj.write)


if __name__ == '__main__':
    main()



"""
Main file for the Bear package.
"""

NAME = 'bear'
PYPI_NAME = 'thebear'
PYPI_INDEX = 'https://pypi.org/simple'
VERSION = '0.4.1'
GITHUB_WEB = 'https://github.com'
GITHUB_API = 'https://api.github.com'
REPO_PATH = 'KeyWeeUsr/Bear'
REPO = f'{GITHUB_WEB}/KeyWeeUsr/Bear'
DESCRIPTION = 'Bear - the decluttering deduplicator'
AUTHOR = 'Peter Badida'
AUTHOR_EMAIL = 'keyweeusr@gmail.com'
LOGO = r"""
 ,--.-'--.,.
/¯  Bear   °¯;
\ , ___ ,\¯ˇ¯
|_;_\ /_;_\
"""
HELP = fr"""
     /¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯\
  <  Use {NAME} -h or {NAME} --help |
    \                            /
       ¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯
"""
LOGO_HELP = '\n'.join([
    left + right
    for left, right in zip(LOGO.split('\n'), HELP.split('\n'))
])



"""
Main module for running the package as a Python module from console:

    python -m <package>
"""

import sys
import logging
from os import stat, remove
from argparse import ArgumentParser, Namespace

from ensure import ensure_annotations

from bear import NAME, LOGO, LOGO_HELP
from bear.common import Hasher
from bear.hashing import hash_file, hash_files
from bear.output import (
    find_files, filter_files, find_duplicates, output_duplicates
)

LOG = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.ERROR,
    format='%(asctime)s - %(process)d - %(name)s - %(levelname)s - %(message)s'
)


@ensure_annotations
def set_log_levels(level: int):
    """
    Set log levels for all available loggers at runtime.
    """
    loggers = [
        logging.getLogger(name)
        for name in logging.root.manager.loggerDict
    ]
    for logger in loggers:
        logger.setLevel(level)


@ensure_annotations
def print_logo():
    """
    Print logo at the beginning of the CLI output.
    """
    if len(sys.argv) == 1:
        print(LOGO_HELP)
    else:
        print(LOGO)


@ensure_annotations
def handle_duplicates(args: Namespace, hasher: Hasher):
    """
    Handle --duplicate related behavior.
    """
    duplicates = find_duplicates(
        folders=args.duplicates, hasher=hasher, processes=args.jobs
    )
    output_duplicates(duplicates, args.output)
    if args.keep_oldest:
        for dups in duplicates.values():
            # oldest == smallest timestamp
            without_oldest = sorted(
                dups, key=lambda item: stat(item).st_mtime
            )[1:]
            for file in without_oldest:
                remove(file)

    elif args.keep_newest:
        for dups in duplicates.values():
            # reverse for oldest
            without_newest = sorted(
                dups, key=lambda item: stat(item).st_mtime, reverse=True
            )[1:]
            for file in without_newest:
                remove(file)


@ensure_annotations
def get_hasher(args: Namespace) -> Hasher:
    """
    Get non-MD5 hasher if desired.
    """
    result = Hasher.MD5  # defalt to MD5
    if args.blake2:
        result = Hasher.BLAKE2
    elif args.sha256:
        result = Hasher.SHA256
    elif args.md5:
        result = Hasher.MD5
    return result


@ensure_annotations
def main(args: Namespace):
    """
    Main function for calling the API from the package depending on
    the CLI options.
    """
    if not args.quiet:
        print_logo()

    if args.quiet:
        set_log_levels(logging.NOTSET)
    elif 0 < args.verbose <= 1:
        set_log_levels(logging.WARNING)
    elif 1 < args.verbose <= 2:
        set_log_levels(logging.INFO)
    elif args.verbose > 2:
        set_log_levels(logging.DEBUG)

    LOG.info('Setting up default logging level to %s', LOG.level)
    LOG.debug('CLI args: %s', args)

    hasher = get_hasher(args)

    if args.files:
        for file in args.files:
            print(hash_file(path=file, hasher=hasher))
    elif args.traverse:
        for folder in args.traverse:
            print(find_files(folder))
    elif args.hash:
        print(filter_files(hash_files(files=[
            file
            for file_list in [find_files(folder) for folder in args.hash]
            for file in file_list
        ], hasher=hasher)))
    elif args.duplicates:
        handle_duplicates(args, hasher=hasher)


class BearArgumentParser(ArgumentParser):
    """
    Custom ArgumentParser to display ASCII logo.
    """

    def print_help(self, file=None):
        print(LOGO)
        super(BearArgumentParser, self).print_help(file)

    def error(self, message):
        print(LOGO)
        super(BearArgumentParser, self).error(message)


def run():
    """
    CLI arguments parser for the main function.
    """
    parser = BearArgumentParser(prog=NAME)
    parser.add_argument('-v', '--verbose', action='count', default=0)
    parser.add_argument(
        '-f', '--files', metavar='FILE', type=str, nargs='+',
        help='files for hashing'
    )
    parser.add_argument(
        '-t', '--traverse', metavar='FOLDER', type=str, nargs='+',
        help='list all files in these folders recursively'
    )
    parser.add_argument(
        '-s', '--hash', metavar='FOLDER', type=str, nargs='+',
        help='hash all files in these folders recursively'
    )
    parser.add_argument(
        '-d', '--duplicates', metavar='FOLDER', type=str, nargs='+',
        help='find all duplicated files in these folders recursively'
    )
    parser.add_argument(
        '-j', '--jobs', action='store', type=int, default=1,
        help='set how many processes will be spawn for hashing, 0=max'
    )
    parser.add_argument(
        '-o', '--output', action='store', type=str, default='',
        help='output file for the list of duplicates'
    )
    parser.add_argument(
        '-q', '--quiet', action='store_true',
        help='suppress all output'
    )
    parser.add_argument(
        '-e', '--keep-oldest', action='store_true',
        help='in combination with --duplicates keep only single oldest file'
    )
    parser.add_argument(
        '-n', '--keep-newest', action='store_true',
        help='in combination with --duplicates keep only single newest file'
    )
    parser.add_argument(
        '--md5', action='store_true',
        help='use MD5 function for hashing (default)'
    )
    parser.add_argument(
        '--blake2', action='store_true',
        help='use BLAKE2 function for hashing'
    )
    parser.add_argument(
        '--sha256', action='store_true',
        help='use SHA256 function for hashing'
    )
    main(parser.parse_args())


if __name__ == '__main__':
    run()



"""
Module for common functions and classes.
"""

from os import getpid
from enum import Enum
from ensure import ensure_annotations


@ensure_annotations
def ignore_append(ignored: str):
    """
    Append a file path to a file with all ignored files.
    """
    with open(f'{getpid()}_ignored.txt', 'a') as out:
        out.write(ignored)
        out.write('\n')


class Hasher(Enum):
    """
    Enum to switch between multiple hashing algorithms.
    """
    MD5 = 1
    SHA256 = 2
    BLAKE2 = 3



"""
Module for output-related functions.
"""
import logging
from os import walk, cpu_count
from os.path import exists, join, abspath, realpath
from multiprocessing import Pool
from datetime import datetime
from functools import partial
from ensure import ensure_annotations

from bear.common import Hasher
from bear.hashing import hash_files

LOG = logging.getLogger(__name__)


@ensure_annotations
def find_files(folder: str) -> list:
    """
    Walk a folder to create a flat list of files.
    """

    result = []

    if not exists(folder):
        LOG.critical('Folder %s does not exist! Skipping.', folder)
        return result

    result = [
        join(name, fname)
        for name, folder, files in walk(folder)
        for fname in files
    ]
    return result


@ensure_annotations
def filter_files(files: dict) -> dict:
    """
    Filter out hashes with only single file connected to them to prevent
    having all of the file hashes even if the files are not duplicated.
    """

    return {key: value for key, value in files.items() if len(value) > 1}


@ensure_annotations
def find_duplicates(folders: list, hasher: Hasher, processes: int = 1) -> dict:
    """
    Find duplicates in multiple folders with multiprocessing.
    """
    # get user specified or max jobs
    processes = processes if processes != 0 else cpu_count()

    # traverse the input folders
    found = [find_files(abspath(realpath(folder))) for folder in folders]
    files = [file for file_list in found for file in file_list]
    files_len = len(files)
    chunk_size = int(files_len // processes)
    if chunk_size == 0:
        # watch out for zero division
        LOG.critical((
            'Zero chunk size for files: %d, processes: %d! '
            'Using 1 as default.'
        ), files_len, processes)
        chunk_size = 1

    # hash chunks of flat list files
    with Pool(processes=processes) as pool:
        results = pool.map(
            # because starmap uses positional args which will become unsafer
            # on each change to the workflow (i.e. more work to find bugs)
            partial(hash_files, hasher=hasher), [
                # chunk files into smaller lists
                files[idx: idx + chunk_size]
                for idx in range(files_len)
                if idx % chunk_size == 0
            ]
        )

    # join values from all jobs
    files = {}
    for result in results:
        for key, val in result.items():
            if key not in files:
                files[key] = val
            else:
                files[key].extend(val)

    # filter out non-duplicates
    return filter_files(files)


@ensure_annotations
def output_duplicates(hashes: dict, out: str = ''):
    """
    Output a simple structure for the duplicates:

    <hash>:
    \t<path>
    \t<path>
    ...
    """
    stamp = str(datetime.now()).replace(':', '_').replace(' ', '_')
    out = f'{stamp}_duplicates.txt' if not out else out

    with open(out, 'wb') as fout:
        for key, val in hashes.items():
            if not isinstance(key, bytes):
                key = key.encode('utf-8', 'ignore')

            # hash
            fout.write(key)
            fout.write(b'\n')

            # tabbed file path(s), exclude invalid characters
            for item in val:
                fout.write(b'\t' + str(item).encode('utf-8', 'ignore') + b'\n')

            # separator
            fout.write(b'\n\n')



#!/usr/bin/env python
"""
Basic setup.py
"""

from os.path import abspath, dirname, join, relpath
from setuptools import setup, find_packages
from bear import (
    VERSION, NAME, REPO, PYPI_NAME, DESCRIPTION, AUTHOR, AUTHOR_EMAIL
)


ROOT = abspath(dirname(__file__))
PKG = join(ROOT, NAME)


with open(join(ROOT, "README.md")) as fd:
    README = fd.read()


DATA = [
    relpath('README.md', ROOT),
    relpath('LICENSE.txt', ROOT)
]


KWARGS = dict(
    name=PYPI_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=REPO,
    download_url=f'{REPO}/zipball/{VERSION}',
    entry_points={
        'console_scripts': [f'{NAME} = {NAME}.__main__:run']
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: '
        'GNU Affero General Public License v3 or later (AGPLv3+)',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Typing :: Typed'
    ],
    install_requires=['ensure'],
    extras_require={
        'dev': ['pycodestyle', 'pylint', 'coverage'],
        'ci': ['coveralls'],
        'release': [
            'setuptools', 'wheel',
            'pycodestyle', 'pylint',
            'coverage', 'coveralls',
            'twine', 'pyinstaller', 'requests'
        ]
    },
    include_package_data=True,
    data_files=[(NAME, DATA)]
)


if __name__ == '__main__':
    setup(**KWARGS)



"""
Test hashing of specific file contents.
"""

from unittest import TestCase, main
from tempfile import mkstemp
from os import remove
from bear.common import Hasher
from bear.hashing import hash_text, hash_file


class HashCase(TestCase):
    """
    Test hashing functions.
    """

    def test_hash_word_md5(self):
        """
        Test hashing a string.
        """
        self.assertEqual(
            '098f6bcd4621d373cade4e832627b4f6',
            hash_text(inp='test'.encode('utf-8'), hasher=Hasher.MD5)
        )

    def test_hash_file_md5(self):
        """
        Test hashing bytes of a file.
        """
        base = b'\xecO\xda\xad\x12\x85\xa1\xe1\xbd\xb7\xf1'
        (desc, path) = mkstemp(text=False)
        with open(desc, 'wb') as file:
            file.write(base)
        hashed = hash_file(path=path, hasher=Hasher.MD5)
        remove(path)
        self.assertEqual('7972892f41d1b98a71d9583b83267d8b', hashed)


if __name__ == '__main__':
    main()



"""
Module for hashing-related functions.
"""
import logging
import traceback
from ensure import ensure_annotations

from bear.common import ignore_append, Hasher

LOG = logging.getLogger(__name__)


@ensure_annotations
def hash_text(inp: bytes, hasher: Hasher) -> str:
    """
    Hash simple string of text.
    """
    result = None
    if hasher == Hasher.MD5:
        from hashlib import md5
        result = md5(inp).hexdigest()
    elif hasher == Hasher.SHA256:
        from hashlib import sha256
        result = sha256(inp).hexdigest()
    elif hasher == Hasher.BLAKE2:
        from hashlib import blake2b
        result = blake2b(inp).hexdigest()
    return result


@ensure_annotations
def hash_file(path: str, hasher: Hasher) -> str:
    """
    Open a file, read its contents and return MD5 hash.
    """
    result = ''
    try:
        with open(path, 'rb') as file:
            contents = file.read()
        result = hash_text(inp=contents, hasher=hasher)
    except PermissionError:
        LOG.critical(
            'Could not open %s due to permission error! %s',
            path, traceback.format_exc()
        )
        ignore_append(path)
    except FileNotFoundError:
        LOG.warning('Could not find file %s! Did you delete it?', path)
        ignore_append(path)
    except OSError:
        LOG.critical('Could not open %s! %s', path, traceback.format_exc())
        ignore_append(path)
    return result


def hash_files(files: list, hasher: Hasher) -> dict:
    """
    Hash each of the file in the list.

    In case of a MemoryError (limitation of e.g. 32-bit Python)
    write out the file names in separate .txt files per PID
    of the process used for hashing.
    """

    hashfiles = {}
    files_len = len(files)

    for idx, fname in enumerate(files):
        LOG.debug('Hashing %d / %d', idx + 1, files_len)

        try:
            fhash = hash_file(path=fname, hasher=hasher)
            if not fhash:
                continue
            if fhash not in hashfiles:
                hashfiles[fhash] = [fname]
            else:
                hashfiles[fhash].extend([fname])
        except MemoryError:
            LOG.critical(
                'Not enough memory while hashing %s, skipping.', fname
            )
            ignore_append(fname)

    return hashfiles
